/*
 *                    BioJava development code
 *
 * This code may be freely distributed and modified under the
 * terms of the GNU Lesser General Public Licence.  This should
 * be distributed with the code.  If you do not have a copy,
 * see:
 *
 *      http://www.gnu.org/copyleft/lesser.html
 *
 * Copyright for this code is held jointly by the individual
 * authors.  These should be listed in @author doc comments.
 *
 * For more information on the BioJava project and its aims,
 * or to join the biojava-l mailing list, visit the home page
 * at:
 *
 *      http://www.biojava.org/
 *
 * Created on 2011-11-20
 *
 */

package org.biojava3.ws.alignment.qblast2;

import static org.biojava3.ws.alignment.qblast2.enums.BlastParameter.DATABASE;
import static org.biojava3.ws.alignment.qblast2.enums.BlastParameter.EXPECT;
import static org.biojava3.ws.alignment.qblast2.enums.BlastParameter.OTHER_ADVANCED;
import static org.biojava3.ws.alignment.qblast2.enums.BlastParameter.PROGRAM;
import static org.biojava3.ws.alignment.qblast2.enums.BlastParameter.QUERY_FROM;
import static org.biojava3.ws.alignment.qblast2.enums.BlastParameter.QUERY_TO;
import static org.biojava3.ws.alignment.qblast2.enums.BlastParameter.WORD_SIZE;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.biojava3.ws.alignment.RemotePairwiseAlignmentProperties;
import org.biojava3.ws.alignment.qblast2.enums.BlastProgram;

/**
 * This class implements {@code RemotePairwiseAlignmentProperties} by adding
 * several convenient methods used to wrap the addition of Blast alignment
 * parameters.
 * <p>
 * Required parameters are PROGRAM and DATABASE, other parameters are optional
 * <p>
 * Any other QBlast URL API parameters can be added using
 * {@link #setAlignementOption(String, String)}
 * 
 * @author Sylvain Foisy, Diploide BioIT
 * @author Gediminas Rimsa (gediminas.rimsa at gmail.com)
 */
public class NCBIQBlastAlignmentProperties2 implements RemotePairwiseAlignmentProperties {


	private static final long serialVersionUID = -7710581294480758153L;

	private Map<String, String> param = new HashMap<String, String>();

	/**
	 * @return the name of blastall program used for blast run
	 */
	public String getBlastProgram() {
		return param.get(PROGRAM);
	}

	/**
	 * Sets the program to be used with blastall
	 * 
	 * @param program : one of blastall programs
	 */
	public void setBlastProgram(BlastProgram program) {
		param.put(PROGRAM, program.getValue());
	}

	/**
	 * @return the name of the database used for blast run
	 */
	public String getBlastDatabase() {
		return param.get(DATABASE);
	}

	/**
	 * Sets the database to be used with blastall
	 * 
	 * @param database : a valid name to a NCBI Blastable database
	 */
	public void setBlastDatabase(String database) {
		param.put(DATABASE, database);
	}

	/**
	 * @return double value of EXPECT parameter used for blast run
	 */
	public double getBlastExpect() {
		if (param.containsKey(EXPECT)) {
			return Double.parseDouble(param.get(EXPECT));
		}
		return 10;
	}

	/**
	 * Sets the EXPECT parameter to be use with blastall
	 * <p>
	 * Example: if you want a EXPECT of 1e-10, pass
	 * {@code Double.parseDouble("1e-10")} as a parameter
	 * 
	 * @param expect: a double value of EXPECT parameter
	 */
	public void setBlastExpect(double expect) {
		param.put(EXPECT, Double.toString(expect));
	}

	/**
	 * Returns the value of the WORD_SIZE parameter used for this blast run
	 * 
	 * @return int value of WORD_SIZE used by this search
	 * @throws IllegalArgumentException when program type is not set or program
	 *             type is not supported
	 */
	public int getBlastWordSize() {
		if (param.containsKey(WORD_SIZE)) {
			return Integer.parseInt(param.get(WORD_SIZE));
		}

		String program = param.get(PROGRAM);
		if (program == null) {
			throw new IllegalArgumentException("Can not return default WORD_SIZE value as blast program has not been set");
		}

		try {
			BlastProgram programType = BlastProgram.valueOf(param.get(PROGRAM).toUpperCase());
			if (BlastProgram.BLASTN.equals(programType)) {
				return 11;
			} else {
				// for blastp / blastx / tblastn / tblastx
				return 3;
			}
			// FIXME: return 28 for megablast when it is added to supported programs enum
		} catch (IllegalArgumentException e) {
			throw new IllegalArgumentException("Blast program " + program + " is not supported.", e);
		}
	}

	/**
	 * Sets the WORD_SIZE parameter to be use with blastall
	 * 
	 * @param expect: an int used to set WORD_SIZE
	 */
	public void setBlastWordSize(int word) {
		param.put(WORD_SIZE, Integer.toString(word));
	}

	/**
	 * Sets the QUERY_FROM and QUERY_TO parameters to be use by blast. Do not
	 * use if you want to use the whole sequence
	 */
	public void setBlastFromToPosition(int start, int end) {
		param.put(QUERY_FROM, String.valueOf(start));
		param.put(QUERY_TO, String.valueOf(end));
	}

	/**
	 * @return QUERY_FROM parameter
	 * @see #setBlastFromToPosition(int, int)
	 */
	public int getBlastFromPosition() {
		return Integer.parseInt(param.get(QUERY_FROM));
	}

	/**
	 * @return QUERY_TO parameter
	 * @see #setBlastFromToPosition(int, int)
	 */
	public int getBlastToPosition() {
		return Integer.parseInt(param.get(QUERY_TO));
	}

	/**
	 * <p>
	 * This method is to be used if a request is to use non-default values at
	 * submission. According to QBlast info, the accepted parameters for PUT
	 * requests are:
	 * </p>
	 * <ul>
	 * <li>-G: cost to create a gap. Default = 5 (nuc-nuc) / 11 (protein) /
	 * non-affine for megablast</li>
	 * <li>-E: Cost to extend a gap. Default = 2 (nuc-nuc) / 1 (protein) /
	 * non-affine for megablast</li>
	 * <li>-r: integer to reward for match. Default = 1</li>
	 * <li>-q: negative integer for penalty to allow mismatch. Default = -3</li>
	 * <li>-y: dropoff for blast extensions in bits, using default if not
	 * specified. Default = 20 for blastn, 7 for all others (except megablast
	 * for which it is not applicable).</li>
	 * <li>-X: X dropoff value for gapped alignment, in bits. Default = 30 for
	 * blastn/megablast, 15 for all others.</li>
	 * <li>-Z: final X dropoff value for gapped alignement, in bits. Default =
	 * 50 for blastn, 25 for all others (except megablast for which it is not
	 * applicable)</li>
	 * <li>-P: equals 0 for multiple hits 1-pass, 1 for single hit 1-pass. Does
	 * not apply to blastn ou megablast.</li>
	 * <li>-A: multiple hits window size. Default = 0 (for single hit algorithm)
	 * </li>
	 * <li>-I: number of database sequences to save hits for. Default = 500</li>
	 * <li>-Y: effective length of the search space. Default = 0 (0 represents
	 * using the whole space)</li>
	 * <li>-z: a real specifying the effective length of the database to use.
	 * Default = 0 (0 represents the real size)</li>
	 * <li>-c: an integer representing pseudocount constant for PSI-BLAST.
	 * Default = 7</li>
	 * <li>-F: any filtering directive</li>
	 * </ul>
	 * <p>
	 * WARNING!! This method is still very much in flux and might not work as
	 * expected...
	 * </p>
	 * <p>
	 * You have to be aware that at no moment is there any error checking on the
	 * use of these parameters by this class.
	 * </p>
	 * 
	 * @param advancedOptions : a String with any number of optional parameters
	 *            with an associated value
	 */
	public void setAdvancedOptions(String advancedOptions) {
		param.put(OTHER_ADVANCED, advancedOptions);
	}

	/**
	 * @return the String with the advanced options
	 */
	public String getBlastAdvancedOptions() {
		return param.get(OTHER_ADVANCED);
	}

	@Override
	public String getAlignmentOption(String key) {
		if (!param.containsKey(key)) {
			throw new IllegalArgumentException("The key named " + key + " is not set for this alignment properties object");
		}
		return param.get(key);
	}

	@Override
	public void setAlignementOption(String key, String val) {
		param.put(key, val);
	}

	@Override
	public Set<String> getAlignmentOptions() {
		return param.keySet();
	}
}
